//------------------------------------------------------------------------------
//  :	Control/main.c 
// :	   ATtiny2313,    
//    PDM. 
//   :
//    1.     ,    
//              -
//    2.     ( PDM) ("" +/-)
//    3.       "."
//    4.        (USART) 
//         ( PDM)
//
//   :
//      :  1 - 1000  ( - 1 )
//      : 0 - 100%    1%.  
//             0    ". ".
//           ""  0.5     
//           10%.
//
//       USART
//        - 8 , 2 -   
//        250 /   16 
//    : 2  8  - , 
//                 HD44780
//
// :	      ATtiny2313
//  :  16
// :		        3.7.6
// :		           .., ku_@mail.ru
//
//  .., 2011
//------------------------------------------------------------------------------

// !!!    ""    
#define START_DISABLE

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

#define RS          PB2
#define E           PB3

#define OffBut      PD2     //      INT0 (PD2)
#define PPlusBut    PD3     //   "+" -  PD3
#define PMinusBut   PD4     //   "-" -  PD4
#define SingleBut   PD6     //    

//--------------------------------------------------------------------------------
#define NoClick           0

#define PPlusShortClick   0b01010000
#define PMinusShortClick  0b01001000
#define SingleShortClick  0b00011000

#define PPlusLongClick    0b00001010
#define PMinusLongClick   0b00001001
#define SingleLongClick   0b00000011

#define FPButsMask        0b01011000
//--------------------------------------------------------------------------------
#define PAUSE 10          //        /
#define PMax 100          //    (100%).  -  0%

char fs[] = {0x20, 0x20, 0x20, 0x2E, 0x20, 0x20, 0x20, 0x6B, 0x00},
     ps[] = {0x50, 0x44, 0x4D, 0x3D, 0x20, 0x20, 0x20, 0x25, 0x00};
     
unsigned long f = 1000000L; //   
unsigned char fReady = 1, cnt = 0;
unsigned int tcnt;

unsigned char rStep = 0;  //       
unsigned char PChangeFlag = 1; // 1,     
unsigned char P;          //    (%)

//------------------------------------------------------------------------------
//  USART
//------------------------------------------------------------------------------
void usart_init(unsigned int ubrr) {
  UBRRH = (unsigned char)(ubrr >> 8);
  UBRRL = (unsigned char)ubrr;
  UCSRB = _BV(TXEN);                //  
  UCSRC = _BV(UCSZ0) | _BV(UCSZ1) | _BV(USBS); //  8  
                                               // 2 -
}
//------------------------------------------------------------------------------
//  
//------------------------------------------------------------------------------
void usart_sb(unsigned char b) {
  while(!(UCSRA & _BV(UDRE)));  //      
  UDR = b;
}
//-----------------------------------------------------------------------------
//   " +/-"
//-----------------------------------------------------------------------------
unsigned char get_key(void){
  unsigned char i, bFlag = PIND & FPButsMask;
  if(bFlag == FPButsMask){ //      
    rStep = 0; return NoClick;
  }
  _delay_ms(PAUSE); //   
  if((PIND & FPButsMask) == FPButsMask){ // - ...
    rStep = 0; return NoClick;
  }
  if(rStep){  //  " "
    if(rStep < 10)  //  
      rStep++;
    _delay_ms(200); // ,     
    return (bFlag >> 3);  //   
  }
  for(i = 0; i < 50; i++){   //      0.5 
    if((PIND & FPButsMask) == FPButsMask) // -   
      return bFlag;
    _delay_ms(10);
  }
  rStep = 2; //  " "
  return (bFlag >> 3);    
}
//-----------------------------------------------------------------------------
//   INT0    
//-----------------------------------------------------------------------------
ISR(INT0_vect){
  P = 0;
  usart_sb(P & 0x7F);
  PChangeFlag = 1;
}//------------------------------------------------------------------------------
//    
//------------------------------------------------------------------------------
void lcd_cmd(unsigned char p) { 	// p -  
	PORTB = (p & 0b11110000) | 
          (PORTB & 0b00001111);	    //    -   PB4-PB7
  PORTB &= ~_BV(RS);                // RS = 0
	PORTB |= _BV(E); _delay_us(100);	// E = 1
	PORTB &= ~_BV(E);	_delay_us(500); // E -> 0 -     

	PORTB = ((p << 4) & 0b11110000) |
          (PORTB & 0b00001111);	    //    -    PB4-PB7
  PORTB &= ~_BV(RS);                // RS = 0
	PORTB |= _BV(E); _delay_us(100);  // E = 1
	PORTB &= ~_BV(E);	_delay_us(500); // E -> 0 -     
}
//------------------------------------------------------------------------------
//    
//------------------------------------------------------------------------------
void lcd_dat(unsigned char p) {		// p -  
	PORTB = (p & 0b11110000) | _BV(RS) |
          (PORTB & 0b00001111);	    //    -   PB4-PB7
  PORTB |= _BV(RS);                // RS = 1
	PORTB |= _BV(E); _delay_us(100);	// E = 1
	PORTB &= ~_BV(E);	_delay_us(500);	// E -> 0 -     

	PORTB = ((p<<4) & 0b11110000) | _BV(RS) |
          (PORTB & 0b00001111);	    //    -   PB4-PB7
  PORTB |= _BV(RS);                // RS = 1
	PORTB |= _BV(E); _delay_us(100);	// E = 1
	PORTB &= ~_BV(E); _delay_us(500);	// E -> 0 -     
}
//------------------------------------------------------------------------------
//  
//------------------------------------------------------------------------------
void ini_cmd(unsigned char p) {   // p -  ,    
	PORTB = (p & 0b11110000) |
          (PORTB & 0b00001111);	          //    -   PB4-PB7
	PORTB &= ~_BV(RS);			          // RS = 0
	PORTB |= _BV(E); _delay_us(100);  // E = 1
	PORTB &= ~_BV(E); _delay_us(500); //     
}
//------------------------------------------------------------------------------
//  
//------------------------------------------------------------------------------
void lcd_init(void) {
	_delay_ms(20); 
	ini_cmd(0x30); _delay_ms(5);
	ini_cmd(0x30); _delay_us(200);
	ini_cmd(0x30); _delay_us(100);
	ini_cmd(0x20); _delay_us(100);
	lcd_cmd(0x28); _delay_us(100);
	lcd_cmd(0x08); _delay_us(100);
	lcd_cmd(0x01); _delay_us(2000);
	lcd_cmd(0x06); _delay_us(100);
	lcd_cmd(0x0D); _delay_us(100);
}
//------------------------------------------------------------------------------
//    
//------------------------------------------------------------------------------
void lcd_str(unsigned char l, char* s) {
	if(l == 0) //    
		lcd_cmd(0x80);
	else      //    ...
		lcd_cmd(0xC0);
	_delay_us(100);
	
	unsigned char i = 0;
	while(s[i] != 0x00)
		lcd_dat(s[i++]);
}
//------------------------------------------------------------------------------
//     TIMER0   A
//    8 , 125     1 
//------------------------------------------------------------------------------
ISR(TIMER0_COMPA_vect){
  tcnt = TCNT1; TCNT1 = 0; cnt++;
  f += tcnt;
  if(cnt == 125){           //  1      
    fReady = 1;             // "" 
    TIMSK &= ~_BV(OCIE0A);  //   TIMER0   A
  }
}
//------------------------------------------------------------------------------
//    1 
//------------------------------------------------------------------------------
void timer_start(void){
  f = 0; fReady = 0; cnt = 0;
  TCNT1 = 0; TCNT0 = 0;     //  
  TIFR |= _BV(OCF0A);       //    (  1)
  TIMSK |= _BV(OCIE0A);     //   TIMER0   A
}
//------------------------------------------------------------------------------
//  
//------------------------------------------------------------------------------
int main (void) {
	unsigned char i, s, d0, d1, d2, d[6]; 

	DDRB = 0xFF; PORTB = 0x01;  //   B -  .

  //    - 1.     
  //   .   (1 )   
  //  ,   TIMER0 (125   8 )
  TCCR1A = 0x00;        // COM1 = 0000 WGM = 0000 -  Normal
  TCCR1B = _BV(CS12)|   // CS = 111 -    ( ) 
           _BV(CS11)|
           _BV(CS10);

  //    - 0.    
  //  8 ,       TIMER1 
  //   ,   
  TCCR0A = _BV(WGM01);  // COM0 = 0000, WGM = 010 -  CTC
  TCCR0B = _BV(CS02)|   // CS = 101 -   
           _BV(CS00);   //   1024
  OCR0A =  124;         // 125     - 8 . 
                        // -1  -    , 
                        //   ,   ..

  GIMSK |= _BV(INT0);	  //    Int0.   
                        //     .
	MCUCR = _BV(ISC01);	  //     

  DDRD &= 0b10000011;   //  2-6  D -  
  PORTD |= 0b01111100;  //   

	lcd_init();           //  
	lcd_cmd(0x0C); _delay_us(100);
	
	usart_init(3);        // 250 /   16

  TIMSK |= _BV(OCIE0A); //       0

  sei();
	while(1) {
    if(fReady){ //    -     LCD
      if(f <1000000L) {
        for(i = 0; i < 6; i++){ //    
          d[i] = f % 10; f /= 10;
        }
        s = 0;
        for(i = 0; i < 6; i++){ //      LCD
          if(i == 3)
            s = 1;
          fs[i + s] = 0x30 + d[5 - i];
        }
        if(d[5] == 0){
          fs[0] = 0x20;
          if(d[4] == 0)
            fs[1] = 0x20;
        }
      } else { //   1  -   
        s = 0;
        for(i = 0; i < 6; i++){ //      LCD
          if(i == 3)
            s = 1;
          fs[i + s] = '*';
        }
      }
      lcd_str(0, fs);
      timer_start();
    }

    switch(get_key()){  //   
      case PPlusShortClick: //     
        if(P < PMax){
          P++; PChangeFlag = 1;
        }
        break;
      case PMinusShortClick: //     
        if(P > 0){
          P--; PChangeFlag = 1;
        }
        break;
      case PPlusLongClick: //     
        if(P <= (PMax - rStep)){
          P += rStep; PChangeFlag = 1;
        }else{
          P = PMax; PChangeFlag = 1;
        }
        break;
      case PMinusLongClick: //     
        if(P >= rStep){
          P -= rStep; PChangeFlag = 1;
        }else{
          P = 0; PChangeFlag = 1;
        }
        break;
      case SingleShortClick: //    ""
      case SingleLongClick: //    ""
#ifndef START_DISABLE
          PORTB |= _BV(PB0);  //      
#endif
          break;
    }

    if(PChangeFlag){
      usart_sb(P & 0x7F);   //     
                            //  PDM
      //    
      d0 = P % 10; //   
      d1 = (P % 100)/10;
      d2 = P/100;
      ps[4] = ((d2 == 0)? 0x20 : (0x30 + d2)); //   0 
      ps[5] = ((d2 == 0) && (d1 == 0))? 0x20 : (0x30 + d1); // ...
      ps[6] = 0x30 + d0;
      lcd_str(1, ps);
      PChangeFlag = 0;
#ifndef START_DISABLE
      if(P == 0)
        PORTB &= ~_BV(PB0); //      
#endif
    }
  }
	return 0;
}
